﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using CQG;

namespace OptionsData
{
   /// <summary>
   /// API facade. Subscribes to instrument properties, monitores CQG connection status.
   /// </summary>
   public class DataAccess : INotifyPropertyChanged, IDisposable
   {
      /// <summary>
      /// API entry object.
      /// </summary>
      private readonly CQGCEL _cel = new();

      /// <summary>
      /// Instrument properties to listen to.
      /// </summary>
      private readonly IEnumerable<eInstrumentProperty> _neededProperties;

      public DataAccess(IEnumerable<eInstrumentProperty> neededProperties)
      {
         _neededProperties = neededProperties;

         _cel.DataConnectionStatusChanged += CEL_DataConnectionStatusChanged;
         _cel.InstrumentResolved += CEL_InstrumentResolved;
         _cel.InstrumentChanged += CEL_InstrumentChanged;
         _cel.DataError += CEL_DataError;
         _cel.APIConfiguration.CollectionsThrowException = false;
         _cel.APIConfiguration.ReadyStatusCheck = eReadyStatusCheck.rscOff;
         _cel.APIConfiguration.NewInstrumentMode = true;
         _cel.APIConfiguration.GatewayEnabled = false;
      }

      /// <summary>
      /// True if connected to CQG client, false otherwise.
      /// </summary>
      public bool IsConnected
      {
         get { return _cel.Environment.DataConnectionStatus == eConnectionStatus.csConnectionUp; }
      }

      /// <summary>
      /// Resolved instruments data.
      /// </summary>
      public ObservableCollection<InstrumentItem> ResolvedInstruments { get; private set; } = new();

      /// <inheritdoc/>
      public event PropertyChangedEventHandler? PropertyChanged;

      /// <summary>
      /// Raised when API error occured.
      /// </summary>
      public event EventHandler<Error>? ErrorOccurred;

      /// <summary>
      /// Connect to API.
      /// </summary>
      public void Start() => _cel.Startup();

      /// <inheritdoc/>
      public void Dispose() => _cel.Shutdown();

      /// <summary>
      /// Sends subscription request for provided symbol.
      /// </summary>
      /// <param name="symbol">Request parameter.</param>
      public void Subscribe(string symbol)
      {
         var request = _cel.CreateInstrumentRequest();
         request.Symbol = symbol;
         request.PropertyLevel = ePropertySubscriptionLevel.psNetChange | ePropertySubscriptionLevel.psGreeks
               | ePropertySubscriptionLevel.psVolatility | ePropertySubscriptionLevel.psTheoreticalValue;
         _cel.SubscribeNewInstrument(request);
      }

      #region Event handlers

      /// <summary>
      /// This event is fired when a new instrument is resolved and subscribed or resolution failed.
      /// </summary>
      /// <param name="symbol">
      /// The commodity symbol that was requested by the user in the NewInstrument method.
      /// </param>
      /// <param name="instrument">
      /// The resolved CQGInstrument object.
      /// </param>
      /// <param name="error">
      /// CQGError instance. It is null in case resolution success.
      /// </param>
      private void CEL_InstrumentResolved(string symbol, CQGInstrument instrument, CQGError error)
      {
         try
         {
            if (_cel.IsValid(error))
            {
               ErrorOccurred?.Invoke(this, new Error(error) { Source = $"CEL_InstrumentResolved '{symbol}'" });
               return;
            }

            if ((instrument.InstrumentType & eInstrumentType.itAllOptions) == 0)
            {
               _cel.RemoveInstrument(instrument);
               var message = $"Requested instrument '{symbol}' resolved into '{instrument.FullName}' which is not an option.";
               ErrorOccurred?.Invoke(this, new Error(message) { Source = "CEL_InstrumentResolved" });
               return;
            }

            for (int i = 0; i < ResolvedInstruments.Count; ++i)
            {
               if (ResolvedInstruments[i].FullName == instrument.FullName)
               {
                  // Replace instrument so UI can track the change and select recently resolved instrument.
                  ResolvedInstruments[i] = new InstrumentItem(_neededProperties, instrument);
                  return;
               }
            }

            ResolvedInstruments.Add(new InstrumentItem(_neededProperties, instrument));
         }
         catch (Exception ex)
         {
            ErrorOccurred?.Invoke(this, new Error(ex) { Source = $"CEL_InstrumentResolved '{symbol}'" });
         }
      }

      /// <summary>
      /// This event is fired when any of the instrument quotes or dynamic instrument properties are changed.
      /// </summary>
      /// <param name="Instrument">
      /// The changed instrument.
      /// </param>
      /// <param name="Quotes">
      /// The collection of changed quotes.
      /// </param>
      /// <param name="props">
      /// The collection of changed dynamic properties.
      /// </param>
      private void CEL_InstrumentChanged(CQGInstrument instrument, CQGQuotes quotes, CQGInstrumentProperties props)
      {
         try
         {
            foreach (var item in ResolvedInstruments)
            {
               if (item.FullName == instrument.FullName)
               {
                  item.Update(props);
               }
            }
         }
         catch (Exception ex)
         {
            ErrorOccurred?.Invoke(this, new Error(ex) { Source = "CEL_InstrumentChanged" });
         }
      }

      /// <summary>
      /// This event is fired, when CQGCEL detects some abnormal discrepancy between data expected and data received.
      /// </summary>
      /// <param name="obj">
      /// Object, in which the error occurred.
      /// </param>
      /// <param name="errorDescription">
      /// String, describing the error.
      /// </param>
      private void CEL_DataError(object obj, string errorDescription)
      {
         try
         {
            if (obj is CQGError error)
            {
               ErrorOccurred?.Invoke(this, new Error(error) { Source = "CEL_DataError" });
            }
            else
            {
               ErrorOccurred?.Invoke(this, new Error(errorDescription) { Source = "CEL_DataError" });
            }
         }
         catch (Exception ex)
         {
            ErrorOccurred?.Invoke(this, new Error(ex) { Source = "CEL_DataError" });
         }
      }

      /// <summary>
      /// This event is fired, when some changes occur in the connection with the CQG data server.
      /// </summary>
      /// <param name="newStatus">
      /// The current status of the connection with the data server.
      /// </param>
      private void CEL_DataConnectionStatusChanged(eConnectionStatus newStatus)
      {
         PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(nameof(IsConnected)));
      }

      #endregion

      #region Nested types

      /// <summary>
      /// <see cref="CQGInstrumentProperty"/> presentation snapshot.
      /// </summary>
      public sealed class PropertyItem : INotifyPropertyChanged
      {
         public eInstrumentProperty Type { get; private set; }
         public string Name { get; private set; } = "";
         public string Value { get; private set; } = "";

         public event PropertyChangedEventHandler? PropertyChanged;

         public PropertyItem(eInstrumentProperty property)
         {
            Type = property;
         }

         public void Update(CQGInstrumentProperty? cqgProperty)
         {
            if (cqgProperty != null && cqgProperty.Type == Type)
            {
               Name = cqgProperty.Name;
               var cel = cqgProperty.Instrument.CEL;
               Value = cel.IsValid(cqgProperty.Value) ? Convert.ToString(cqgProperty.Value) : "N/A";
               PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(nameof(Value)));
            }
         }
      }

      /// <summary>
      /// <see cref="CQGInstrument"/> presentation snapshot.
      /// </summary>
      public sealed class InstrumentItem
      {
         public string FullName { get; private set; }

         public IEnumerable<PropertyItem> Properties { get; private set; }

         public InstrumentItem(IEnumerable<eInstrumentProperty> neededProperties, CQGInstrument cqgInstrument)
         {
            FullName = cqgInstrument.FullName;
            Properties = neededProperties.Select(prop => new PropertyItem(prop)).ToArray();
            foreach (var item in Properties)
            {
               item.Update(cqgInstrument.Properties[item.Type]);
            }
         }

         public void Update(CQGInstrumentProperties cqgProps)
         {
            foreach (CQGInstrumentProperty cqgProp in cqgProps)
            {
               Properties.FirstOrDefault(property => property.Type == cqgProp.Type)?.Update(cqgProp);
            }
         }
      }

      /// <summary>
      /// API Error.
      /// </summary>
      public class Error
      {
         /// <summary>
         /// API Error code.
         /// https://partners.cqg.com/sites/default/files/docs/CQGAPI_4.0R/webframe.html#Error%20Codes.html
         /// </summary>
         public int Code { get; private set; }

         /// <summary>
         /// Error message.
         /// </summary>
         public string Message { get; private set; }

         /// <summary>
         /// Place where the error occurred.
         /// </summary>
         public string? Source { get; set; }

         public Error(CQGError error)
         {
            Code = error.Code;
            Message = error.Description;
         }

         public Error(Exception ex)
         {
            Message = ex.Message;
         }

         public Error(string message)
         {
            Message = message;
         }
      }

      #endregion
   }
}
