﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Globalization;
using System.Windows;
using System.Windows.Data;
using System.Linq;
using CQG;

namespace OptionsData
{
   /// <summary>
   /// Interaction logic for MainWindow.xaml
   /// </summary>
   public partial class MainWindow : Window
   {
      /// <summary>
      /// Option properties to show.
      /// </summary>
      private static readonly IEnumerable<eInstrumentProperty> NeededProperties = new[]
      {
         eInstrumentProperty.ipDescription,
         eInstrumentProperty.ipStrike,
         eInstrumentProperty.ipExpirationDate,
         eInstrumentProperty.ipNetChange,
         eInstrumentProperty.ipDelta,
         eInstrumentProperty.ipDeltaNetChange,
         eInstrumentProperty.ipTheta,
         eInstrumentProperty.ipThetaNetChange,
         eInstrumentProperty.ipGamma,
         eInstrumentProperty.ipGammaNetChange,
         eInstrumentProperty.ipVega,
         eInstrumentProperty.ipVegaNetChange,
         eInstrumentProperty.ipRho,
         eInstrumentProperty.ipRhoNetChange,
         eInstrumentProperty.ipAverageVolatility,
         eInstrumentProperty.ipImpliedVolatility,
         eInstrumentProperty.ipImpliedVolatilityOnBid,
         eInstrumentProperty.ipImpliedVolatilityNetChange,
         eInstrumentProperty.ipVolatilityCurve,
         eInstrumentProperty.ipTheoreticalValue,
         eInstrumentProperty.ipTheoValueNetChange
      };

      /// <summary>
      /// Bound to option data grid as FallbackValue.
      /// </summary>
      public static readonly IEnumerable<DataAccess.PropertyItem> EmptyProperties = 
         NeededProperties.Select(prop => new DataAccess.PropertyItem(prop)).ToArray();

      /// <summary>
      /// CQG API facade. Window data context.
      /// </summary>
      private readonly DataAccess _dataAccess = new(NeededProperties);

      /// <summary>
      /// Window construction and initialization.
      /// </summary>
      public MainWindow()
      {
         InitializeComponent();
         DataContext = _dataAccess;
         _dataAccess.ResolvedInstruments.CollectionChanged += ResolvedInstruments_CollectionChanged;
         _dataAccess.ErrorOccurred += DataAccess_ErrorOccurred;
      }

      /// <summary>
      /// Handle API errors.
      /// </summary>
      private void DataAccess_ErrorOccurred(object? sender, DataAccess.Error e)
      {
         string title = e.Source ?? "Error.";
         if (e.Code == 103 || e.Code == 125)
         {
            title = "Turn on CQG Client and restart the application.";
         }
            
         MessageBox.Show(e.Message, title, MessageBoxButton.OK, MessageBoxImage.Error);
      }

      /// <summary>
      /// Select recently resolved instrument.
      /// </summary>
      private void ResolvedInstruments_CollectionChanged(object? sender, NotifyCollectionChangedEventArgs e)
      {
         if (e.NewItems != null && e.NewItems.Count > 0)
         {
            lvResolvedContracts.SelectedItem = e.NewItems[0];
         }
      }

      /// <summary>
      /// Start event.
      /// </summary>
      private void Window_Loaded(object sender, RoutedEventArgs e)
      {
         try
         {
            _dataAccess.Start();
         }
         catch (Exception ex)
         {
            MessageBox.Show($"Failed to start CEL: {ex.Message}", "Initialization Error",
                MessageBoxButton.OK, MessageBoxImage.Error);
            Close();
         }
      }

      /// <summary>
      /// Exit event.
      /// </summary>
      private void Window_Closed(object sender, System.EventArgs e)
      {
         try
         {
            _dataAccess.Dispose();
         }
         catch
         {
            // window is closed, just ignore it
         }
      }

      /// <summary>
      /// Handles input event in "Symbol" text box.
      /// </summary>
      private void TextBoxSymbol_KeyUp(object sender, System.Windows.Input.KeyEventArgs e)
      {
         if (e.Key == System.Windows.Input.Key.Enter && TextBoxSymbol.Text.Trim().Length > 0)
         {
            BtnSubscribe_Click(sender, e);
         }
      }

      /// <summary>
      /// Handles "Subscribe" button click event.
      /// </summary>
      private void BtnSubscribe_Click(object sender, RoutedEventArgs e)
      {
         try
         {
            _dataAccess.Subscribe(TextBoxSymbol.Text.Trim());
         }
         catch (Exception ex)
         {
            MessageBox.Show($"Error occurred: {ex.Message}", "Subscribe", MessageBoxButton.OK, MessageBoxImage.Error);
         }
         TextBoxSymbol.Text = "";
      }
   }

   /// <summary>
   /// Converts <see cref="DataAccess.IsConnected"/> to main window title text.
   /// </summary>
   public class WindowTitleConverter : IValueConverter
   {
      public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
      {
         return (value is bool connected && connected) ? "OptionsData [Connected]" : "OptionsData [Connection is DOWN]";
      }

      public object ConvertBack(object value, Type targetType, object parameter, CultureInfo culture)
      {
         throw new NotImplementedException();
      }
   }
}
